import logging
import numbers
from datetime import date, datetime
from decimal import Decimal
from uuid import UUID

import six

log = logging.getLogger("usermaven")


def remove_trailing_slash(host):
    if host.endswith("/"):
        return host[:-1]
    return host


def clean(item):
    if isinstance(item, Decimal):
        return float(item)
    if isinstance(item, UUID):
        return str(item)
    elif isinstance(item, (six.string_types, bool, numbers.Number, datetime, date, type(None))):
        return item
    elif isinstance(item, (set, list, tuple)):
        return _clean_list(item)
    elif isinstance(item, dict):
        return _clean_dict(item)
    else:
        return _coerce_unicode(item)


def _clean_list(list_):
    return [clean(item) for item in list_]


def _clean_dict(dict_):
    data = {}
    for k, v in six.iteritems(dict_):
        try:
            data[k] = clean(v)
        except TypeError:
            log.warning(
                'Dictionary values must be serializeable to JSON "%s" value %s of type %s is unsupported.',
                k,
                v,
                type(v),
            )
    return data


def _coerce_unicode(cmplx):
    try:
        item = cmplx.decode("utf-8", "strict")
    except AttributeError as exception:
        item = ":".join(exception)
        item.decode("utf-8", "strict")
        log.warning("Error decoding: %s", item)
        return None
    return item




import json
import logging
from datetime import date, datetime
from typing import Any, Optional, Union

import requests

from usermaven.utils import remove_trailing_slash
from usermaven.settings import DEFAULT_HOST, USER_AGENT

_session = requests.sessions.Session()


def post(
    api_key: str, server_token: str, host: Optional[str] = None, path=None, timeout: int = 15, **kwargs
) -> requests.Response:
    """Post the `kwargs` to the API"""
    log = logging.getLogger("usermaven")
    body = kwargs
    url = remove_trailing_slash(host or DEFAULT_HOST) + path
    data = body["batch"]
    log.debug("making request: %s", data)
    headers = {"Content-Type": "application/json", "User-Agent": USER_AGENT}
    server_secret_key = api_key + "." + server_token
    res = _session.post(url, params={'token': server_secret_key}, json=data, headers=headers, timeout=timeout)

    if res.status_code == 200:
        log.debug("data uploaded successfully")

    return res


def _process_response(
    res: requests.Response, success_message: str, *, return_json: bool = True
) -> Union[requests.Response, Any]:
    log = logging.getLogger("usermaven")
    if res.status_code == 200:
        log.debug(success_message)
        return res.json() if return_json else res
    try:
        payload = res.json()
        log.debug("received response: %s", payload)
        raise APIError(res.status_code, payload["detail"])
    except (KeyError, ValueError):
        raise APIError(res.status_code, res.text)


def batch_post(
    api_key: str, server_token: str, host: Optional[str] = None, timeout: int = 15, **kwargs
) -> requests.Response:
    """Post the `kwargs` to the batch API endpoint for events"""
    res = post(api_key, server_token, host, "/api/v1/s2s/event/", timeout, **kwargs)
    return _process_response(res, success_message="data uploaded successfully", return_json=False)


class APIError(Exception):
    def __init__(self, status: Union[int, str], message: str):
        self.message = message
        self.status = status

    def __str__(self):
        msg = "[Usermaven] {0} ({1})"
        return msg.format(self.message, self.status)


class DatetimeSerializer(json.JSONEncoder):
    def default(self, obj: Any):
        if isinstance(obj, (date, datetime)):
            return obj.isoformat()

        return json.JSONEncoder.default(self, obj)



import atexit
import logging
import random
import string

from six import string_types

from usermaven.consumer import Consumer
from usermaven.request import batch_post
from usermaven.utils import clean
from usermaven.settings import ID_TYPES

try:
    import queue
except ImportError:
    import Queue as queue


class Client(object):
    """Create a new Usermaven client."""

    log = logging.getLogger("usermaven")

    def __init__(
        self,
        api_key=None,
        server_token=None,
        host=None,
        debug=False,
        max_queue_size=10000,
        send=True,
        on_error=None,
        flush_at=100,
        flush_interval=0.5,
        max_retries=3,
        sync_mode=False,
        timeout=15,
        thread=1,
    ):

        self.queue = queue.Queue(max_queue_size)

        # api_key: This is the project_id/workspace_id which is required for authentication
        self.api_key = stringify_id(api_key)

        # server_token: This should be the server secret token used for authentication when sending the events.
        self.server_token = stringify_id(server_token)

        require("api_key", self.api_key, string_types)
        require("server_token", self.server_token, string_types)

        self.on_error = on_error
        self.debug = debug
        self.send = send
        self.sync_mode = sync_mode
        self.host = host
        self.timeout = timeout
        self.group_type_mapping = None

        if debug:
            # Ensures that debug level messages are logged when debug mode is on.
            # Otherwise, defaults to WARNING level. See https://docs.python.org/3/howto/logging.html#what-happens-if-no-configuration-is-provided
            logging.basicConfig()
            self.log.setLevel(logging.DEBUG)
        else:
            self.log.setLevel(logging.WARNING)

        if sync_mode:
            self.consumers = None
        else:
            # On program exit, allow the consumer thread to exit cleanly.
            # This prevents exceptions and a messy shutdown when the
            # interpreter is destroyed before the daemon thread finishes
            # execution. However, it is *not* the same as flushing the queue!
            # To guarantee all messages have been delivered, you'll still need
            # to call flush().
            if send:
                atexit.register(self.join)
            for n in range(thread):
                self.consumers = []
                consumer = Consumer(
                    self.queue,
                    self.api_key,
                    self.server_token,
                    host=host,
                    on_error=on_error,
                    flush_at=flush_at,
                    flush_interval=flush_interval,
                    retries=max_retries,
                    timeout=timeout,
                )
                self.consumers.append(consumer)

                # if we've disabled sending, just don't start the consumer
                if send:
                    consumer.start()

    def identify(self, user, company={}):
        require("user", user, dict)
        if "id" in user and "email" in user and "created_at" in user:
            # user object has required attributes
            require("user_id", user["id"], ID_TYPES)
            require("user_email", user["email"], string_types)
            require("user_created_at", user["created_at"], string_types)
        else:
            # user object is missing one or more of the required attributes
            raise ValueError("user object is missing one or more of the required attributes")

        msg = {
            "api_key": self.api_key,
            "event_id": "",
            "event_type": "user_identify",
            "ids": {},
            "user": {
                "anonymous_id": generate_id(),
                "id": user["id"],
                "email": user["email"],
                "created_at": user["created_at"],
            },
            "screen_resolution": "0",
            "src": "usermaven-python"
        }

        if "custom" in user:
            require("user_custom", user["custom"], dict)
            msg["user"]["custom"] = user["custom"]

        if company:
            require("company", company, dict)
            if "id" in company and "name" in company and "created_at" in company:
                # company object has required attributes
                require("company_id", company["id"], ID_TYPES)
                require("company_name", company["name"], string_types)
                require("company_created_at", company["created_at"], string_types)
            else:
                # company object is missing one or more of the required attributes
                raise ValueError("company object is missing one or more of the required attributes")

            msg["company"] = {
                "id": company["id"],
                "name": company["name"],
                "created_at": company["created_at"]
            }
            if "custom" in company:
                require("company_custom", company["custom"], dict)
                msg["company"]["custom"] = company["custom"]

        return self._enqueue(msg)

    def track(self, user_id, event_type, company={}, event_attributes={}):
        require("user_id", user_id, ID_TYPES)
        require("event_type", event_type, string_types)

        msg = {
            "api_key": self.api_key,
            "event_type": event_type,
            "event_id": "",
            "ids": {},
            "user": {
                "anonymous_id": generate_id(),
                "id": user_id
            },
            "screen_resolution": "0",
            "src": "usermaven-python",
            "event_attributes": event_attributes
        }

        if company:
            require("company", company, dict)
            if "id" in company and "name" in company and "created_at" in company:
                # company object has required attributes
                require("company_id", company["id"], ID_TYPES)
                require("company_name", company["name"], string_types)
                require("company_created_at", company["created_at"], string_types)
            else:
                # company object is missing one or more of the required attributes
                raise ValueError("company object is missing one or more of the required attributes")

            msg["company"] = {
                "id": company["id"],
                "name": company["name"],
                "created_at": company["created_at"]
            }
            if "custom" in company:
                require("company_custom", company["custom"], dict)
                msg["company"]["custom"] = company["custom"]

        return self._enqueue(msg)

    def _enqueue(self, msg):
        """Push a new `msg` onto the queue, return `(success, msg)`"""

        msg = clean(msg)
        self.log.debug("queueing: %s", msg)

        # if send is False, return msg as if it was successfully queued
        if not self.send:
            return True, msg

        if self.sync_mode:
            self.log.debug("enqueued with blocking %s.", msg["event_type"])
            batch_post(self.api_key, self.server_token, self.host, timeout=self.timeout, batch=[msg])

            return True, msg

        try:
            self.queue.put(msg, block=False)
            self.log.debug("enqueued %s.", msg["event_type"])
            return True, msg
        except queue.Full:
            self.log.warning("analytics-python queue is full")
            return False, msg

    def flush(self):
        """Forces a flush from the internal queue to the server"""
        queue = self.queue
        size = queue.qsize()
        queue.join()
        # Note that this message may not be precise, because of threading.
        self.log.debug("successfully flushed about %s items.", size)

    def join(self):
        """Ends the consumer thread once the queue is empty.
        Blocks execution until finished
        """
        for consumer in self.consumers:
            consumer.pause()
            try:
                consumer.join()
            except RuntimeError:
                # consumer thread has not started
                pass

    def shutdown(self):
        """Flush all messages and cleanly shutdown the client"""
        self.flush()
        self.join()


def require(name, field, data_type):
    """Require that the named `field` has the right `data_type`"""
    if not isinstance(field, data_type):
        msg = "{0} must have {1}, got: {2}".format(name, data_type, field)
        raise AssertionError(msg)


def stringify_id(val):
    if val is None:
        return None
    if isinstance(val, string_types):
        return val
    return str(val)


def generate_id():
    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=10))



from typing import Callable, Dict, Optional

from usermaven.client import Client
from usermaven.version import VERSION

__version__ = VERSION

"""Settings."""
api_key = None  # type: str
server_token = None  # type: str
host = None  # type: str
on_error = None  # type: Callable
debug = False  # type: bool
send = True  # type: bool
sync_mode = False  # type: bool
disabled = False  # type: bool

default_client = None


def track(
    user_id,              # type: str
    event_type,           # type: str
    company={},           # type: Optional[Dict]
    event_attributes={},  # type: Optional[Dict]
):
    # type: (...) -> None
    """
    Track allows you to track anything a user does within your system, which you can later use in Usermaven
    to find patterns in usage, work out which features to improve or where people are giving up.

    A `track` call requires
    - `user_id` which is the unique identifier for the user.
    - `event_type` which defines what sort of event are you tracking from this call. We recommend using [verb] [noun],
     like `goal created` or `payment succeeded` to easily identify what your events mean later on.

    Optionally you can submit
    - `company`, which is a dict with company properties. Name, id and created_at are required fields for the
    company object. You can also submit custom properties for the company object.
    - 'event_attributes', which is a dict that contain information about the event.


    For example:
    ```python
    usermaven.track('user_id', 'payment_succeeded')
    usermaven.track('user_id', 'video_watched', 'event_attributes'={'video_title': 'demo', 'watched_at': '2022'})
    usermaven.track('user_id', 'plan_purchased', company={'name': 'Company Name', 'id': '1', 'created_at': '2022',
    'custom': {'plan': 'enterprise', 'industry': 'Technology'}}, event_attributes={'plan': 'premium'})
    ```
    """
    _proxy(
        "track",
        user_id=user_id,
        event_type=event_type,
        company=company,
        event_attributes=event_attributes
    )


def identify(
    user,        # type: Dict
    company={},  # type: Optional[Dict]
):
    # type: (...) -> None
    """
    Identify lets you add metadata to the user so you can know who they are in Usermaven.

    An `identify` call requires
    - `user` which is a dict of user properties. Email, id and created_at are required fields for the user object.
    You can also submit custom properties for the user object.

    For example:
    ```python
    usermaven.identify({'email': 'john@gmail.com','id': '123', 'created_at': '2022'})
    ```
    Optionally you can submit
    - `company`, which is a dict with company properties. Name, id and created_at are required fields for the
    company object. You can also submit custom properties for the company object.

    """
    _proxy(
        "identify",
        user=user,
        company=company
    )


def flush():
    """Tell the client to flush."""
    _proxy("flush")


def join():
    """Block program until the client clears the queue"""
    _proxy("join")


def shutdown():
    """Flush all messages and cleanly shutdown the client"""
    _proxy("flush")
    _proxy("join")


def _proxy(method, *args, **kwargs):
    """Create an analytics client if one doesn't exist and send to it."""
    global default_client
    if disabled:
        return None
    if not default_client:
        default_client = Client(
            api_key,
            server_token,
            host=host,
            debug=debug,
            on_error=on_error,
            send=send,
            sync_mode=sync_mode,
        )

    fn = getattr(default_client, method)
    return fn(*args, **kwargs)



from six import string_types
import numbers
from uuid import UUID
from usermaven.version import VERSION


ID_TYPES = (numbers.Number, string_types, UUID)

MAX_MSG_SIZE = 32 << 10
# Our servers only accept batches less than 500KB. Here limit is set slightly
# lower to leave space for extra data that will be added later, eg. "sentAt".
BATCH_SIZE_LIMIT = 475000

DEFAULT_HOST = "https://eventcollectors.usermaven.com"
USER_AGENT = "usermaven-python/" + VERSION



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['usermaven', 'usermaven.test']

package_data = \
{'': ['*'], 'usermaven': ['assets/images/logos/*']}

install_requires = \
['backoff>=1.10.0,<2.0.0',
 'monotonic>=1.5',
 'python-dateutil>2.1',
 'requests>=2.7,<3.0',
 'six>=1.5']

setup_kwargs = {
    'name': 'usermaven',
    'version': '0.1.1',
    'description': '',
    'long_description': '<p align="center">\n  <a href="https://usermaven.com/">\n    <picture>\n        <source media="(prefers-color-scheme: dark)" srcset="https://storage.googleapis.com/lumotive-web-storage/usermaven/usermaven-white.png" height="60">\n        <img alt="Dark/Light mode logos" src="https://storage.googleapis.com/lumotive-web-storage/usermaven/usermaven-dark.webp" height="60">\n    </picture>\n  </a>\n  <p align="center">PRIVACY-FRIENDLY ANALYTICS TOOL</p>\n</p>\n\n![PyPI](https://img.shields.io/badge/pypi-v0.1.1-blue)\n![Software License](https://img.shields.io/badge/license-MIT-green)\n![PyPI - Python Version](https://img.shields.io/badge/python-3.6%20and%20above-blue)\n\n# Usermaven Python \n\nThis module is compatible with Python 3.6 and above.\n\n## Installing\n\n```bash\npip3 install usermaven\n```\n\n## Usage\n\n```python\nfrom usermaven import Client\nclient = Client(api_key=\'your_workspace_api_key\', server_token="your_workspace_server_token")\nclient.identify(user={\'email\': \'user@domain.com\',\'id\': \'lzL24K3kYw\', \'created_at\': \'2022-01-20T09:55:35\'})\nclient.track(user_id=\'lzL24K3kYw\', event_type=\'signed_up\')\n```\n\n### Instantiating usermaven-python\'s client object\n\nCreate an instance of the client with your Usermaven workspace credentials.\n\n```python\nfrom usermaven import Client\nclient = Client(api_key=\'your_workspace_api_key\', server_token="your_workspace_server_token")\n```\n\n### Create a Usermaven user profile\n\n```python\nclient.identify(user={\'email\': \'user@domain.com\',\'id\': \'lzL24K3kYw\', \'created_at\': \'2022-01-20T09:55:35\'})\n```\n\n#### Required arguments\n`user`: The user object is the only required argument for `identify` call. `email`, `id` and `created_at` are required\nfields for the user object. Recommended fields for the user object are `first_name` and `last_name`. Additionally you \ncan pass any custom properties in the form of dictionary to your user object.\n\n#### Optional arguments\nYou can also pass optional arguments to the `identify` method.\n\n`company`: A company object for which the user belongs to. It is optional but if it is passed, it must contain `name`,\n`id` and `created_at` fields. You can also submit custom properties in the form of dictionary for the company object. \nExample:\n```python\nclient.identify(user={\n    # Required attributes of user object\n    \'email\': \'user@domain.com\', \'id\': \'lzL24K3kYw\', \'created_at\': \'2022-01-20T09:55:35\',\n    \n    # Recommended attributes - First name and last name are shown on people pages.\n    \'first_name\': \'John\', \'last_name\': \'Smith\',\n    \n    # Optional attributes (you can name attributes what you wish)\n    \'custom\': {\'plan_name\': \'premium\'}\n    }, \n    \n    # If your product is used by multiple users in a company, we recommend to pass the Company attributes.\n    company={\n        # Required attributes of company object\n        \'name\': \'Usermaven\', \'id\': \'uPq9oUGrIt\', \'created_at\': \'2022-01-20T09:55:35\',\n        \n        # Optional attributes such as industry, website, employee count etc.\n        \'custom\': {\'plan\': \'enterprise\', \'industry\': \'Technology\', \'website\': \'https://usermaven.com\', \n                   \'employees\': \'20\'}\n    })\n```\n\n### Track a custom event\n\n```python\nclient.track(user_id=\'lzL24K3kYw\', event_type=\'plan_purchased\')\n```\n\n#### Required arguments\n`user_id`: For the `track` call, you must pass the `user_id` of the user you want to track the event for.\n\n`event_type`: For track call, `event_type` is a required argument and you must pass a value to the event_type.\nWe recommend using [verb] [noun], like `goal created` or `payment succeeded` to easily identify what your events mean\nlater on.\n\n#### Optional arguments\nYou can also pass optional arguments to the `track` method.\n\n`company`: A company object for which the user belongs to. It is optional but if it is passed, it must contain `name`,\n`id` and `created_at` fields. You can also submit custom properties in the form of dictionary for the company object.\n\n`event_attributes`: This can contain information related to the event that is being tracked. Example:\n```python\nclient.track(\n    user_id=\'lzL24K3kYw\', # Unique ID for the user in database. (Required)\n    event_type=\'plan_purchased\', # Event name to be tracked (Required)\n    \n    # Optional attributes\n    company={\n        # Required attributes of company object\n        \'name\': \'Usermaven\', \'id\': \'uPq9oUGrIt\', \'created_at\': \'2022-01-20T09:55:35\',\n        \n        # Optional attributes such as industry, website, employee count etc.\n        \'custom\': {\'plan\': \'enterprise\', \'industry\': \'Technology\', \'website\': \'https://usermaven.com\', \n                   \'employees\': \'20\'}\n        },\n    \n    event_attributes={\n        \'plan_name\': \'premium\',\n        \'plan_price\': \'100\',\n        \'plan_currency\': \'USD\'\n        }\n    )\n```\n\n## Local Setup for Development\nFor local development, you can clone the repository and install the dependencies using the following commands:\n\n```bash\ngit clone "https://github.com/usermaven/usermaven-python.git"\npoetry install\n```\n\n## Running tests\n\nChanges to the library can be tested by running `python -m unittest -v` from the parent directory.\n',
    'author': 'usermaven',
    'author_email': 'azhar@usermaven.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import time
import unittest

import mock
import six

from usermaven.client import Client
from usermaven.test.test_utils import FAKE_TEST_SERVER_TOKEN, FAKE_TEST_API_KEY


class TestClient(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # This ensures no real HTTP POST requests are made
        cls.client_post_patcher = mock.patch("usermaven.client.batch_post")
        cls.consumer_post_patcher = mock.patch("usermaven.consumer.batch_post")
        cls.client_post_patcher.start()
        cls.consumer_post_patcher.start()

    @classmethod
    def tearDownClass(cls):
        cls.client_post_patcher.stop()
        cls.consumer_post_patcher.stop()

    def set_fail(self, e, batch):
        """Mark the failure handler"""
        print("FAIL", e, batch)
        self.failed = True

    def setUp(self):
        self.failed = False
        self.client = Client(FAKE_TEST_API_KEY, FAKE_TEST_SERVER_TOKEN, on_error=self.set_fail)
        self.user = {"id": "user_id", "email": "test_user@d4interactive.io", "created_at": "2022-12-12T19:11:49"}
        self.user_id = "user_id"

    def test_requires_api_key(self):
        with self.assertRaises(AssertionError):
            Client(server_token=FAKE_TEST_SERVER_TOKEN)

    def test_requires_server_token(self):
        with self.assertRaises(AssertionError):
            Client(api_key=FAKE_TEST_API_KEY)

    def test_empty_flush(self):
        self.client.flush()

    def test_basic_track(self):
        client = self.client
        success, msg = client.track(self.user_id, "goal_created")
        client.flush()
        self.assertTrue(success)
        self.assertFalse(self.failed)

        self.assertEqual(msg["user"]["id"], "user_id")
        self.assertEqual(msg["api_key"], "random_api_key_for_testing")
        self.assertEqual(msg["event_type"], "goal_created")

    def test_basic_track_without_event_type(self):
        # Test that track method raises a TypeError when called without event_type argument
        with self.assertRaises(TypeError):
            self.client.track(self.user_id)  # missing event_type argument

    def test_advanced_track(self):
        client = self.client
        success, msg = client.track(
            self.user_id,
            "goal_created",
            {"id": "5", "name": "Usermaven", "created_at": "2022-12-12T19:11:49"},
            {"goal_name": "signup", "goal_value": 100},
        )

        self.assertTrue(success)

        self.assertEqual(msg["company"]["name"], "Usermaven")
        self.assertEqual(msg["company"]["id"], "5")
        self.assertEqual(msg["company"]["created_at"], "2022-12-12T19:11:49")
        self.assertEqual(msg["user"]["id"], "user_id")
        self.assertEqual(msg["api_key"], "random_api_key_for_testing")
        self.assertEqual(msg["event_type"], "goal_created")
        self.assertEqual(msg["event_attributes"]["goal_name"], "signup")
        self.assertEqual(msg["event_attributes"]["goal_value"], 100)

    def test_company_track_without_name(self):
        # Test that track method raises a ValueError when called with company dictionary missing name key
        with self.assertRaises(ValueError):
            self.client.track(self.user_id, "goal_created", company={"id": "5", "created_at": "2022-12-12T19:11:49"})

    def test_company_track_with_custom(self):
        client = self.client
        success, msg = client.track(
            self.user_id,
            "goal_created",
            {"id": "5", "name": "Usermaven", "created_at": "2022-12-12T19:11:49",
             "custom": {"custom_key": "custom_value"}},
        )

        self.assertTrue(success)
        self.assertEqual(msg["company"]["custom"]["custom_key"], "custom_value")

    def test_basic_identify(self):
        client = self.client
        success, msg = client.identify(self.user)
        client.flush()
        self.assertTrue(success)
        self.assertFalse(self.failed)

        self.assertEqual(msg["user"]["id"], "user_id")
        self.assertEqual(msg["user"]["email"], "test_user@d4interactive.io")
        self.assertEqual(msg["user"]["created_at"], "2022-12-12T19:11:49")
        self.assertEqual(msg["api_key"], "random_api_key_for_testing")
        self.assertEqual(msg["event_type"], "user_identify")
        self.assertEqual(msg["src"], "usermaven-python")

    def test_advanced_identify(self):
        client = self.client
        success, msg = client.identify(
            self.user,
            company={"id": "5", "name": "Usermaven", "created_at": "2022-12-12T19:11:49"}
        )

        self.assertTrue(success)

        self.assertEqual(msg["company"]["name"], "Usermaven")
        self.assertEqual(msg["company"]["id"], "5")
        self.assertEqual(msg["company"]["created_at"], "2022-12-12T19:11:49")
        self.assertEqual(msg["src"], "usermaven-python")
        self.assertEqual(msg["user"]["id"], "user_id")
        self.assertEqual(msg["user"]["email"], "test_user@d4interactive.io")
        self.assertEqual(msg["user"]["created_at"], "2022-12-12T19:11:49")
        self.assertEqual(msg["event_id"], "")
        self.assertEqual(msg["api_key"], "random_api_key_for_testing")
        self.assertEqual(msg["event_type"], "user_identify")

    def test_user_identify_without_email(self):
        # Test that identify method raises a ValueError when called with dictionary missing email key
        with self.assertRaises(ValueError):
            self.client.identify({"id": "user_id", "created_at": "2022-12-12T19:11:49"})

    def test_company_identify_without_id(self):
        # Test that identify method raises a ValueError when called with company dictionary missing id key
        with self.assertRaises(ValueError):
            self.client.identify(self.user, company={"name": "Usermaven"})

    def test_advanced_identify_with_custom(self):
        client = self.client
        success, msg = client.identify(
            {"id": "user_id", "email": "test_user@d4interactive.io", "created_at": "2022-12-12T19:11:49",
             "custom": {"custom_key": "custom_value"}},
            {"id": "5", "name": "Usermaven", "created_at": "2022-12-12T19:11:49",
             "custom": {"custom_key": "custom_value"}},
        )

        self.assertTrue(success)
        self.assertEqual(msg["company"]["custom"]["custom_key"], "custom_value")
        self.assertEqual(msg["user"]["custom"]["custom_key"], "custom_value")

    def test_flush(self):
        client = self.client
        # set up the consumer with more requests than a single batch will allow
        for i in range(1000):
            success, msg = client.identify(self.user)
        # We can't reliably assert that the queue is non-empty here; that's
        # a race condition. We do our best to load it up though.
        client.flush()
        # Make sure that the client queue is empty after flushing
        self.assertTrue(client.queue.empty())

    def test_shutdown(self):
        client = self.client
        # set up the consumer with more requests than a single batch will allow
        for i in range(1000):
            success, msg = client.identify(self.user)
        client.shutdown()
        # we expect two things after shutdown:
        # 1. client queue is empty
        # 2. consumer thread has stopped
        self.assertTrue(client.queue.empty())
        for consumer in client.consumers:
            self.assertFalse(consumer.is_alive())

    def test_synchronous(self):
        client = Client(FAKE_TEST_API_KEY, FAKE_TEST_SERVER_TOKEN, sync_mode=True)

        success, message = client.identify(self.user)
        self.assertFalse(client.consumers)
        self.assertTrue(client.queue.empty())
        self.assertTrue(success)

    def test_overflow(self):
        client = Client(FAKE_TEST_API_KEY, FAKE_TEST_SERVER_TOKEN, max_queue_size=1)
        # Ensure consumer thread is no longer uploading
        client.join()

        for i in range(10):
            client.identify(self.user)

        success, msg = client.identify(self.user)
        # Make sure we are informed that the queue is at capacity
        self.assertFalse(success)

    def test_unicode(self):
        Client(six.u("unicode_key"), six.u("unicode_token"))

    def test_debug(self):
        Client("bad_key", "bad_token", debug=True)

    def test_user_defined_flush_at(self):
        client = Client(FAKE_TEST_API_KEY, FAKE_TEST_SERVER_TOKEN, on_error=self.fail, flush_at=10, flush_interval=3)

        def mock_post_fn(*args, **kwargs):
            self.assertEqual(len(kwargs["batch"]), 10)

        # the post function should be called 2 times, with a batch size of 10
        # each time.
        with mock.patch("usermaven.consumer.batch_post", side_effect=mock_post_fn) as mock_post:
            for _ in range(20):
                client.identify(self.user)
            time.sleep(1)
            self.assertEqual(mock_post.call_count, 2)

    def test_user_defined_timeout(self):
        client = Client(FAKE_TEST_API_KEY, FAKE_TEST_SERVER_TOKEN, timeout=10)
        for consumer in client.consumers:
            self.assertEqual(consumer.timeout, 10)

    def test_default_timeout_15(self):
        client = Client(FAKE_TEST_API_KEY, FAKE_TEST_SERVER_TOKEN)
        for consumer in client.consumers:
            self.assertEqual(consumer.timeout, 15)
